from typing import Any, Callable, Dict, Generic, Mapping, Optional, Tuple, TypeVar, Union

from chex import Numeric

from ..annotations import PyTree
from ..dataclass import dataclass, fields
from .meta_parameter import MetaParameter

__all__ = ['GradientTransformation', 'GradientState']


GradientState = PyTree
Weights = TypeVar('Weights', bound=PyTree)
State = TypeVar('State', bound=PyTree)


@dataclass
class GradientTransformation(Generic[State, Weights]):

    T = TypeVar('T', bound='GradientTransformation[State, Weights]')

    def init(self, parameters: Weights) -> State:
        raise NotImplementedError

    def update(self,
               gradient: Weights,
               state: State,
               parameters: Optional[Weights]) -> Tuple[Weights, State]:
        raise NotImplementedError

    def has_meta_parameters(self) -> bool:
        return any(isinstance(getattr(self, field.name), MetaParameter)
                   for field in fields(self))

    def replace_meta_parameters(self: T,
                                f: Callable[[MetaParameter], Union[Numeric, MetaParameter]]) -> T:
        replacements: Dict[str, Numeric] = {}
        for field in fields(self):
            name = field.name
            value = getattr(self, name)
            if isinstance(value, MetaParameter):
                replacements[name] = f(value)
        return self.replace(**replacements)

    def replace_meta_parameters_with_defaults(
            self: T,
            meta_parameters: Optional[Mapping[Any, Numeric]] = None) -> T:
        if meta_parameters is None:
            return self

        def f(meta_parameter: MetaParameter) -> Numeric:
            # https://github.com/python/mypy/issues/2608
            return meta_parameters[meta_parameter.key]  # type: ignore
        return self.replace_meta_parameters(f)



from typing import Tuple, TypeVar

from jax.experimental.host_callback import id_print

from .annotations import PyTree
from .shims import custom_vjp

__all__ = ['copy_cotangent', 'print_cotangent']


X = TypeVar('X', bound=PyTree)


@custom_vjp
def copy_cotangent(x: X, y: X) -> Tuple[X, X]:
    return x, y


def _copy_cotangent_fwd(x: X, y: X) -> Tuple[Tuple[X, X], None]:
    return (x, y), None


def _copy_cotangent_bwd(residuals: None, xy_bar: Tuple[X, X]) -> Tuple[X, X]:
    del residuals
    x_bar, _ = xy_bar
    return x_bar, x_bar


copy_cotangent.defvjp(_copy_cotangent_fwd, _copy_cotangent_bwd)


@custom_vjp
def print_cotangent(x: X) -> X:
    return x


def _print_cotangent_fwd(x: X) -> Tuple[X, None]:
    return x, None


def _print_cotangent_bwd(residuals: None, x_bar: X) -> Tuple[X]:
    del residuals
    x_bar = id_print(x_bar)
    return x_bar,


print_cotangent.defvjp(_print_cotangent_fwd, _print_cotangent_bwd)



from __future__ import annotations

from typing import Any, List, Optional, Tuple

import jax.random
import numpy as np
from chex import Array
from jax import numpy as jnp

from .annotations import RealArray, Shape, ShapeLike
from .dataclass import dataclass

__all__ = ['Generator']


@dataclass
class Generator:
    """
    This class represents a JAX random number generator.  Unlike `numpy.Generator`, `tjax.Generator`
    has no mutating methods.  Instead, its generation methods return a new instance along with the
    generated tensor.
    """

    key: Array

    def __init__(self,
                 *,
                 seed: Optional[int] = None,
                 key: Optional[Array] = None,
                 **kwargs: Any):
        super().__init__(**kwargs)
        if key is None:
            if seed is None:
                raise ValueError
            key = jax.random.PRNGKey(seed)
        object.__setattr__(self, 'key', key)

    # New methods ----------------------------------------------------------------------------------
    def split(self, n: int = 2) -> List[Generator]:
        """
        Split a generator into a list of generators.
        """
        assert self.key.shape == (2,)
        keys = jax.random.split(self.key, n)
        return [Generator(key=key) for key in keys]

    def vmap_split(self, shape: ShapeLike) -> Generator:
        """
        Split a generator into a generator that can be passed to a vmapped function.
        """
        if isinstance(shape, int):
            shape = (shape,)
        else:
            shape = tuple(shape)
        prod_shape = np.prod(shape)
        keys = (self.key
                if prod_shape == 1
                else jax.random.split(self.key, prod_shape))
        return Generator(key=keys.reshape(shape + (2,)))

    def bernoulli(self, p: RealArray, shape: Shape = ()) -> Tuple[Generator, Array]:
        g1, g2 = self.split()
        return g1, jax.random.bernoulli(g2.key, p, shape)

    def gamma(self, gamma_shape: RealArray, shape: Shape = ()) -> Tuple[Generator, Array]:
        g1, g2 = self.split()
        return g1, jax.random.gamma(g2.key, gamma_shape, shape)

    def normal(self, std_dev: RealArray, shape: Shape = ()) -> Tuple[Generator, Array]:
        g1, g2 = self.split()
        return g1, std_dev * jax.random.normal(g2.key, shape)

    def uniform(self, minval: RealArray = 0.0, maxval: RealArray = 0.0, shape: Shape = ()) -> (
            Tuple[Generator, Array]):
        g1, g2 = self.split()
        return g1, jax.random.uniform(g2.key, minval=minval, maxval=maxval, shape=shape)

    # Magic methods --------------------------------------------------------------------------------
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Generator):
            return NotImplemented
        return jnp.all(self.key == other.key)

    def __ne__(self, other: Any) -> bool:
        if not isinstance(other, Generator):
            return NotImplemented
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash((int(self.key[0]), int(self.key[1])))



from functools import singledispatch
from numbers import Number
from typing import Any, Dict, List, Tuple, Type, Union

import colorful as cf
import networkx as nx
import numpy as np
from jax.interpreters.ad import JVPTracer
from jax.interpreters.batching import BatchTracer
from jax.interpreters.partial_eval import JaxprTracer
from jax.interpreters.xla import DeviceArray

__all__ = ['print_generic', 'display_generic']


def print_generic(*args: Any, **kwargs: Any) -> None:
    for value in args:
        print(display_generic(value))
    for key, value in kwargs.items():
        print(display_key_and_value(key, value, "=", True, 0))


@singledispatch
def display_generic(value: Any, show_values: bool = True, indent: int = 0) -> str:
    return cf.red(str(value)) + "\n"


@display_generic.register
def _(value: JVPTracer, show_values: bool = True, indent: int = 0) -> str:
    return cf.magenta(f"JVPTracer {value.shape} {value.dtype}") + "\n"


@display_generic.register
def _(value: JaxprTracer, show_values: bool = True, indent: int = 0) -> str:
    return cf.magenta(f"JaxprTracer {value.shape} {value.dtype}") + "\n"


@display_generic.register
def _(value: BatchTracer, show_values: bool = True, indent: int = 0) -> str:
    return cf.magenta(f"BatchTracer {value.shape} {value.dtype} "
                      f"batched over {value.val.shape[value.batch_dim]}") + "\n"


@display_generic.register
def _(value: np.ndarray, show_values: bool = True, indent: int = 0) -> str:
    retval = cf.yellow(f"NumPy Array {value.shape}") + "\n"
    return retval + _show_array(indent + 1, value)


@display_generic.register
def _(value: DeviceArray, show_values: bool = True, indent: int = 0) -> str:
    retval = cf.violet(f"Jax Array {value.shape}") + "\n"
    return retval + _show_array(indent + 1, value)


@display_generic.register(type(None))
@display_generic.register(Number)
def _(value: Union[None, Number], show_values: bool = True, indent: int = 0) -> str:
    return cf.cyan(str(value)) + "\n"


@display_generic.register(dict)
def _(value: Dict[Any, Any], show_values: bool = True, indent: int = 0) -> str:
    return (display_class(dict)
            + "".join(display_key_and_value(key, sub_value, "=", show_values, indent)
                      for key, sub_value in sorted(value.items())))


@display_generic.register(tuple)
@display_generic.register(list)
def _(value: Union[Tuple[Any, ...], List[Any]], show_values: bool = True, indent: int = 0) -> str:
    return (display_class(type(value))
            + "".join(display_key_and_value("", sub_value, "", show_values, indent)
                      for sub_value in value))


@display_generic.register
def _(value: nx.Graph, show_values: bool, indent: int = 0) -> str:
    directed = isinstance(value, nx.DiGraph)
    arrow = cf.base00('⟶  ' if directed else '🡘 ')
    retval = display_class(type(value))
    for name, node in value.nodes.items():
        retval += display_key_and_value(name, node, ": ", show_values, indent)
    for (source, target), edge in value.edges.items():
        key = f"{source}{arrow}{target}"
        retval += display_key_and_value(key, edge, ": ", show_values, indent)
    return retval


def display_class(cls: Type[Any]) -> str:
    return cf.orange(cls.__name__) + "\n"


def display_key_and_value(
        key: str, value: Any, separator: str, show_values: bool = True, indent: int = 0) -> str:
    return (_indent_space(indent + 1) + cf.blue(key) + cf.base00(separator)
            + display_generic(value, show_values, indent + 1))


# Private functions --------------------------------------------------------------------------------
_indentation = 4


def _indent_space(indent: int) -> str:
    return (indent * _indentation) * " "


def _format_number(x: float) -> str:
    return f"{x:10.4f}"


def _show_array(indent: int, array: Union[np.ndarray, DeviceArray]) -> str:
    if len(array.shape) == 0:
        return _indent_space(indent) + _format_number(array[()]) + "\n"
    if np.prod(array.shape) == 0:
        return ""
    if 1 in array.shape:
        return _show_array(indent, array[tuple(0 if s == 1 else slice(None)
                                               for s in array.shape)])
    if len(array.shape) == 1:
        return _indent_space(indent) + "  ".join(_format_number(array[i])
                                                 for i in range(array.shape[0])) + "\n"
    if len(array.shape) == 2:
        return "".join(_show_array(indent, array_slice)
                       for array_slice in array)
    return ""
